unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, db, sqldb, odbcconn, Forms, Controls, Graphics, Dialogs,
  StdCtrls, DBGrids;

type

  { TForm1 }

  TForm1 = class(TForm)
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    DataSource1: TDataSource;
    DBGrid1: TDBGrid;
    Edit1: TEdit;
    Label1: TLabel;
    ODBCConnection1: TODBCConnection;
    SQLQuery1: TSQLQuery;
    SQLTransaction1: TSQLTransaction;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure SQLQuery1AfterOpen(DataSet: TDataSet);
  private
    procedure TextFromDB(Sender: TField; var AText: string; DisplayText: BOOLEAN);
    procedure TextToDB(Sender: TField; const AText: string);

  public

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

uses
  LCLType, LConvEncoding;

const
  MDB_NAME = 'test.mdb';

Const
   ODBC_ADD_DSN=1;
   ODBC_CONFIG_DSN=2;
   ODBC_REMOVE_DSN=3;
   ODBC_ADD_SYS_DSN=4;
   ODBC_CONFIG_SYS_DSN=5;
   ODBC_REMOVE_SYS_DSN=6;
   ODBC_REMOVE_DEFAULT_DSN=7;

function SQLConfigDataSource(hwndParent: Integer; fRequest: Integer;
  lpszDriverString: PChar; lpszAttributes: PChar): Integer; stdcall;
  external 'odbccp32.dll';

function SQLInstallerError(iError: integer; pfErrorCode: PInteger;
  lpszErrorMsg: string; cbErrorMsgMax: integer; pcbErrorMsg: PInteger): integer; stdcall;
  external 'odbccp32.dll';


function CreateAccessDatabase(DatabaseFile: string): boolean;
var
  DBPChar: PChar;
  Driver: PChar;
  ErrorCode, ResizeErrorMessage: integer;
  ErrorMessage: PChar;
  retCode: integer;
begin
//  driver := 'Microsoft Access Driver (*.mdb)';
  driver := 'Microsoft Access Driver (*.mdb, *.accdb)';
  DBPChar:=PChar('CREATE_DBV4="'+DatabaseFile+'"');
  retCode := SQLConfigDataSource(Hwnd(nil), ODBC_ADD_DSN, Driver, DBPChar);
  if retCode<>0 then
  begin
    //try alternate driver
    Driver := 'Microsoft Access Driver (*.md, *.accdb)';
    DBPChar:=PChar('CREATE_DB="'+DatabaseFile+'"');
    retCode := SQLConfigDataSource(Hwnd(nil), ODBC_ADD_DSN, Driver, DBPChar);
  end;
  if retCode=0 then
  begin
    result:=true;
  end
  else
  begin
    result:=false;
    ErrorCode:=0;
    ResizeErrorMessage:=0;
    // todo: verify how the DLL is called - use pointers?; has not been tested.
    GetMem(ErrorMessage,512);
    try
      SQLInstallerError(1, @ErrorCode, ErrorMessage, SizeOf(ErrorMessage), @ResizeErrorMessage);
    finally
      FreeMem(ErrorMessage);
    end;
    raise Exception.CreateFmt('Error creating Access database: %s', [ErrorMessage]);
  end;
end;


{ TForm1 }

procedure TForm1.Button1Click(Sender: TObject);
var
  fn: String;
begin
  if not ODBCConnection1.Connected then begin
    fn := ExtractFilePath(Application.ExeName) + MDB_NAME;
    ODBCConnection1.Driver := 'Microsoft Access Driver (*.mdb, *.accdb)';
    ODBCConnection1.Params.Add('DBQ=' + fn);
    ODBCConnection1.Params.Add('Locale Identifier=1033');
    ODBCConnection1.Params.Add('ExtendedAnsiSQL=1');
    ODBCConnection1.Params.Add('CHARSET=ansi');
    ODBCConnection1.Connected := True;
    ODBCConnection1.KeepConnection := True;

    //transaction
    SQLTransaction1.DataBase := ODBCConnection1;
    SQLTransaction1.Action := caCommit;
    SQLTransaction1.Active := True; //not applied to ms access(false)
  end;

  SQLQuery1.Close;
  SQLQuery1.DataBase := ODBCConnection1;
  SQLQuery1.UsePrimaryKeyAsKey := False;
  SQLQuery1.SQL.Text := UTF8ToCP1252('SELECT * FROM  Prüfung');
  SQLQuery1.Open;
end;

procedure TForm1.Button2Click(Sender: TObject);
var
  fn: String;
begin
  fn := ExtractFilePath(Application.ExeName) + MDB_NAME;
  if FileExists(fn) then DeleteFile(fn);

  CreateAccessDatabase(MDB_NAME);

  ODBCConnection1.Driver := 'Microsoft Access Driver (*.mdb)'; //, *.accdb)';
  ODBCConnection1.Params.Add('DBQ=' + fn);
  ODBCConnection1.Params.Add('Locale Identifier=1033');
  ODBCConnection1.Params.Add('ExtendedAnsiSQL=1');
  ODBCConnection1.Params.Add('CHARSET=ansi');

  SQLTransaction1.StartTransaction;
  ODBCConnection1.ExecuteDirect(UTF8ToCP1252(
    'CREATE TABLE Prüfung ( '+
       'PrüfNr INTEGER, ' +
       'PrüfName VARCHAR(100) )'
  ));
  SQLTransaction1.Commit;

  SQLTransaction1.StartTransaction;
  ODBCConnection1.ExecuteDirect(UTF8ToCP1252(
    'INSERT INTO Prüfung (PrüfNr, PrüfName) VALUES (1, ''ABC'')'));
  ODBCConnection1.ExecuteDirect(UTF8ToCP1252(
    'INSERT INTO Prüfung (PrüfNr, PrüfName) VALUES (2, ''Äöü'')'));
  ODBCConnection1.ExecuteDirect(UTF8ToCP1252(
    'INSERT INTO Prüfung (PrüfNr, PrüfName) VALUES (3, ''Aöü'')'));
  SQLTransaction1.Commit;

  SQLQuery1.UsePrimaryKeyAsKey := False;
  ODBCConnection1.Open;
  SQLQuery1.SQL.Text := UTF8ToCP1252('SELECT * FROM Prüfung');
  SQLQuery1.Open;
end;

procedure TForm1.Button3Click(Sender: TObject);
begin
  SQLQuery1.Close;
  if Edit1.Text = '' then
    SQLQuery1.SQL.Text := UTF8ToCP1252('SELECT * FROM Prüfung')
  else begin
    if SQLQuery1.Params.Count = 0 then
      with TParam(SQLQuery1.Params.Add) do begin
        Name := 'PruefName';
        DataType := ftString;
        ParamType := ptInput;
      end;
    SQLQuery1.ParamByName('PruefName').AsString := '%' + UTF8ToCP1252(Edit1.Text) + '%';
    SQLQuery1.SQL.Text := UTF8ToCP1252('SELECT * FROM Prüfung WHERE PrüfName LIKE :PruefName');
  end;
  SQLQuery1.Open;
end;

procedure TForm1.SQLQuery1AfterOpen(DataSet: TDataSet);
var
  i: Integer;
begin
  for i:=0 to SQLQuery1.FieldCount-1 do
    if SQLQuery1.Fields[i].DataType = ftString then
    begin
      SQLQuery1.Fields[i].OnGetText := @TextFromDB;
      SQLQuery1.Fields[i].OnSetText := @TextToDB;
    end;
  for i:= 0 to DBGrid1.Columns.Count-1 do
    DBGrid1.Columns[i].Title.Caption := CP1252ToUTF8(DBGrid1.Columns[i].Field.FieldName);
end;

procedure TForm1.TextFromDB(Sender: TField; var AText: string;
  DisplayText: BOOLEAN);
begin
  if not Sender.IsNull then
    AText := CP1252ToUTF8(Sender.AsString);
end;

procedure TForm1.TextToDB(Sender: TField; const AText: string);
begin
  if AText <> '' then
    Sender.Value := UTF8ToCP1252(AText);
end;

end.

